from pywr.recorders import Recorder
import time
import logging
logger = logging.getLogger(__name__)


class ProgressRecorder(Recorder):
    """Simple text-based progress notifications

    Parameters
    ----------
    print_func : callable or None
        The function to call when updating progress. The function is given a single str argument that
        contains the progress message. Defaults to `logger.info`.
    """
    def __init__(self, *args, **kwargs):
        print_func = kwargs.pop('print_func', None)
        super(ProgressRecorder, self).__init__(*args, **kwargs)
        if print_func is None:
            print_func = logger.info
        self.print_func = print_func
        self.last_progress = None
        self.last_timestep = None
        self.t0 = None
        self.combinations = None

    def reset(self):
        self.last_progress = -1
        self.last_timestep = 0
        self.t0 = time.time()
        self.combinations = len(self.model.scenarios.combinations)

    def after(self):
        total_timesteps = len(self.model.timestepper)
        timestep = self.model.timestepper.current.index
        progress = int((timestep+1) / total_timesteps * 100)
        if progress > self.last_progress:
            self.last_progress = progress
            if progress >= 1:
                time_taken = time.time() - self.t0
                try:
                    speed = ((timestep-self.last_timestep)*self.combinations) / time_taken
                except ZeroDivisionError:
                    speed = float("inf")
                self.last_timestep = timestep
                self.update_progress(progress, speed)
            else:
                self.last_timestep = timestep
                self.update_progress(progress, None)
            self.t0 = time.time()
            self.last_timestep = timestep

    def update_progress(self, progress, speed=None):
        if speed is not None:
            self.print_func("Completed {}%, {:.0f} steps/second".format(progress, speed))
        else:
            self.print_func("Completed {}%".format(progress))

class JupyterProgressRecorder(ProgressRecorder):
    """Graphical progress bar for use in Jupyter notebooks"""
    def __init__(self, *args, **kwargs):
        super(JupyterProgressRecorder, self).__init__(*args, **kwargs)

    def reset(self):
        from ipywidgets import FloatProgress, HBox, Label, Layout
        from IPython.display import display
        super(JupyterProgressRecorder, self).reset()
        self.progress_bar = FloatProgress(min=0, max=100, description='Running:')
        self.label = Label("", layout=Layout(width='100%'))
        self.box = HBox([self.progress_bar, self.label])
        display(self.box)

    def update_progress(self, progress, speed):
        self.progress_bar.value = progress
        if speed is not None:
            self.label.value = "{:.0f} steps/second".format(speed)
        else:
            self.label.value = ""

    def finish(self):
        super(JupyterProgressRecorder, self).finish()
        if self.progress_bar.value >= 100.0:
            self.progress_bar.bar_style = "success"
        else:
            self.progress_bar.bar_style = "danger"




import numpy as np
import scipy.stats
import pandas
import matplotlib
import matplotlib.pyplot as plt

c = {
    "Afill": "#ee1111",
    "Aedge": "#660000",
    "Bfill": "#1111ee",
    "Bedge": "#000066",
    "Cfill": "#11bb11",
    "Cedge": "#008800",
}

def align_series(A, B, names=None, start=None, end=None):
    """Align two series for plotting / comparison

    Parameters
    ----------
    A : `pandas.Series`
    B : `pandas.Series`
    names : list of strings
    start : `pandas.Timestamp` or timestamp string
    end : `pandas.Timestamp` or timestamp string

    Example
    -------
    >>> A, B = align_series(A, B, ["Pywr", "Aquator"], start="1920-01-01", end="1929-12-31")
    >>> plot_standard1(A, B)

    """
    # join series B to series A
    # TODO: better handling of heterogeneous frequencies
    df = pandas.concat([A, B], join="inner", axis=1)

    # apply names
    if names is not None:
        df.columns = names
    else:
        names = list(df.columns)

    # clip start and end to user-specified dates
    idx = [df.index[0], df.index[-1]]
    if start is not None:
        idx[0] = pandas.Timestamp(start)
    if end is not None:
        idx[1] = pandas.Timestamp(end)

    if start or end:
        df = df.loc[idx[0]:idx[-1],:]

    A = df[names[0]]
    B = df[names[1]]
    return A, B

def plot_standard1(A, B):
    fig, axarr = plt.subplots(3, figsize=(10, 12), facecolor="white")
    plot_timeseries(A, B, axarr[0])
    plot_QQ(A, B, axarr[1])
    plot_percentiles(A, B, axarr[2])
    return fig, axarr

def set_000formatter(axis):
    axis.set_major_formatter(
        matplotlib.ticker.FuncFormatter(lambda x, p: format(int(x), ',')))

def plot_timeseries(A, B, ax=None):
    if ax is None:
        ax = plt.gca()
    B.plot(ax=ax, color=c["Bfill"], clip_on=False)
    A.plot(ax=ax, color=c["Afill"], clip_on=False)
    ax.grid(True)
    ax.set_ylim(0, None)
    set_000formatter(ax.get_yaxis())
    ax.set_xlabel("")
    ax.legend([B.name, A.name], loc="best")
    return ax

def plot_QQ(A, B, ax=None):
    if ax is None:
        ax = plt.gca()
    ax.scatter(B.values, A.values, color=c["Cfill"], edgecolor=c["Cedge"], clip_on=False)
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    limit = max(xlim[1], ylim[0])
    ax.plot([0, limit], [0, limit], '-k')
    ax.set_xlim(0, limit)
    ax.set_ylim(0, limit)
    ax.grid(True)
    set_000formatter(ax.get_xaxis())
    set_000formatter(ax.get_yaxis())
    ax.set_xlabel(B.name)
    ax.set_ylabel(A.name)
    ax.legend(["Equality"], loc="best")
    return ax

def plot_percentiles(A, B, ax=None):
    if ax is None:
        ax = plt.gca()
    percentiles = np.linspace(0.001, 0.999, 1000) * 100
    A_pct = scipy.stats.scoreatpercentile(A.values, percentiles)
    B_pct = scipy.stats.scoreatpercentile(B.values, percentiles)
    percentiles = percentiles / 100.0
    ax.plot(percentiles, B_pct[::-1], color=c["Bfill"], clip_on=False, linewidth=2)
    ax.plot(percentiles, A_pct[::-1], color=c["Afill"], clip_on=False, linewidth=2)
    ax.set_xlabel("Cumulative frequency")
    ax.grid(True)
    ax.xaxis.grid(True, which="both")
    set_000formatter(ax.get_yaxis())
    ax.set_xscale("logit")
    xticks = ax.get_xticks()
    xticks_minr = ax.get_xticks(minor=True)
    ax.set_xticklabels([], minor=True)
    ax.set_xticks([0.01, 0.1, 0.5, 0.9, 0.99])
    ax.set_xticklabels(["1", "10", "50", "90", "99"])
    ax.set_xlim(0.001, 0.999)
    ax.legend([B.name, A.name], loc="best")
    return ax



# -*- coding: utf-8 -*-
from ._parameters import Parameter
from ..core import Storage
import numpy as np


class KeatingStreamFlowParameter(Parameter):
    """
    A flow Parameter that returns stream flow from an Aquifer based on the groundwater level.

    The approach is based on the lumped parameter model by Keating (1982). This Parameter calculates the
     perennial and winterbourne stream flow components based on the Storage node. It requires the Storage
     node to have a valid level Parameter.

    In contrast to the Keating paper a general coefficient is provided for calibration. The Keating approach
     utilised a coefficient of $B/L$ where "$B$ and $L$ are the dimensions of the aquifer block parallel
     and perpendicular to the stream."

    Keating, T. (1982), A Lumped Parameter Model of a Chalk Aquifer-Stream System in Hampshire,
      United Kingdom. Ground Water, 20: 430–436. doi:10.1111/j.1745-6584.1982.tb02763.x
    """
    def __init__(self, model, storage_node, levels, transmissivity, coefficient=1.0, **kwargs):
        """

        :param storage_node:
        :param levels:
        :param transmissivity:
        :param coefficient:
        """
        super(KeatingStreamFlowParameter, self).__init__(model, **kwargs)
        self.storage_node = storage_node

        if len(levels) != len(transmissivity):
            raise ValueError('The number of transmissivity values must equal the number of levels.')

        self.levels = np.array(levels)
        self.transmissivity = np.array(transmissivity)
        self.coefficient = coefficient

    def value(self, ts, scenario_index):
        # Get the current level of the aquifer
        # TODO: this is a HACK - we can't use get_level/get_value as there is
        #       no way to define the parent/child relationship
        level = self.storage_node.level.value(ts, scenario_index)

        # Coefficient
        C = self.coefficient

        # Calculate flow at each stream level
        Q = 0.0
        for n, stream_flow_level in enumerate(self.levels):
            T = self.transmissivity[n]
            if level > stream_flow_level:
                Q += 2 * T * C * (level - stream_flow_level)

        return Q





cdef class Component:
    cdef object _name
    cdef readonly object model
    cdef public str comment
    cdef public dict tags
    cdef readonly object parents
    cdef readonly object children
    cpdef setup(self)
    cpdef reset(self)
    cpdef before(self)
    cpdef after(self)
    cpdef finish(self)
